from tuoni.TuoniExceptions import *

class TuoniListener:
    """
    A class that provides data and functionality for a created listener.

    Attributes:
        listener_id (int): The unique identifier of the listener.
        name (str): The name of the listener.
        info (str): Information about the listener.
        status (str): The current status of the listener.
        plugin (str): The plugin associated with the listener.
        configuration (dict): The configuration settings for the listener.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the listener class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.listener_id = conf["id"]
        self.name = conf["name"]
        self.info = conf["info"]
        self.status = conf["status"]
        self.plugin = conf["plugin"]
        self.configuration = conf["configuration"]

    def stop(self):
        """
        Stop the listener.
        """
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_put(f"/api/v1/listeners/{self.listener_id}/stop")
        self._load_conf(data)

    def start(self):
        """
        Start the listener.
        """
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_put(f"/api/v1/listeners/{self.listener_id}/start")
        self._load_conf(data)

    def delete(self):
        """
        Delete the listener.
        """
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        self.c2.request_delete(f"/api/v1/listeners/{self.listener_id}")
        self.listener_id = None

    def reload(self):
        """
        Reload the listener data from the C2 server.
        """
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/listeners/{self.listener_id}")
        self._load_conf(data)

    def update(self,):
        """
        Update listener on the server.
        """
        if self.listener_id is None:
            raise ExceptionTuoniDeleted("")
        req = {
            "configuration": self.configuration,
            "name": self.name
        }
        self.c2.request_patch(f"/api/v1/listeners/{self.listener_id}", req)
        self.reload()





import time
from tuoni.TuoniExceptions import *


class TuoniDataService:
    """
    A class that provides data and functionality for a service data model entry.

    Attributes:
        id (GUID): The unique identifier (GUID) for the service.
        address (str): The service address.
        port (int): Port number.
        protocol (str): Service protocol.
        banner (str): Service banner.
        note (str): Additional notes.
        status (str): Status of the entry.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the service data class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.id = conf["id"]
        self.address = conf["address"]
        self.port = conf["port"]
        self.protocol = conf["protocol"]
        self.banner = conf["banner"]
        self.note = conf["note"]
        self.status = conf["status"]

    def reload(self):
        """
        Reload the service data from the C2 server.
        """
        if self.id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/discovery/service/{self.id}")
        self._load_conf(data)

    def archive(self):
        """
        Archive the service.
        """
        if self.id is None:
            raise ExceptionTuoniDeleted("")
        self.c2.request_post("/api/v1/discovery/services/bulk-archive", {"serviceIds": [self.id]})

    def update(self):
        """
        Update data on the server.
        """
        if self.id is None:
            raise ExceptionTuoniDeleted("")
        req = {
            "protocol": self.protocol,
            "banner": self.banner,
            "note": self.note
        }
        self.c2.request_patch(f"/api/v1/discovery/service/{self.id}", req)
        self.reload()




import time
from tuoni.TuoniExceptions import *


class TuoniAlias:
    """
    A class that provides data and functionality for a command alias.

    Attributes:
        alias_id (GUID): The unique identifier (GUID) for the alias.
        name (str): The name of the alias.
        qualifiedName (str): The qualified name of the alias.
        fullyQualifiedName (str): The fully qualified name of the alias.
        description (str): A description of the alias.
        fixedConfiguration (dict): The fixed configuration settings for the base command.
        pluginVersion (str): The version of the associated plugin.
        baseTemplate (dict): The base template used for the alias.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the alias class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.alias_id = conf["id"]
        self.name = conf["name"]
        self.qualifiedName = conf["qualifiedName"]
        self.fullyQualifiedName = conf["fullyQualifiedName"]
        self.description = conf["description"]
        self.fixedConfiguration = conf["fixedConfiguration"]
        self.pluginVersion = conf["pluginVersion"]
        self.baseTemplate = conf["baseTemplate"]

    def reload(self):
        """
        Reload the alias data from the C2 server.
        """
        if self.alias_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/command-alias/{self.alias_id}")
        self._load_conf(data)

    def delete(self):
        """
        Delete the alias.
        """
        if self.alias_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_delete(f"/api/v1/command-alias/{self.alias_id}")
        self._load_conf(data)




import os
import time
import requests
import json
import threading
import shutil
import base64
import inspect
from pathlib import Path
from tuoni.TuoniExceptions import *
from tuoni.TuoniListenerPlugin import *
from tuoni.TuoniListener import *
from tuoni.TuoniAgent import *
from tuoni.TuoniPayloadPlugin import *
from tuoni.TuoniAlias import *
from tuoni.TuoniUser import *
from tuoni.TuoniFile import *
from tuoni.TuoniDataHost import *
from tuoni.TuoniDataService import *
from tuoni.TuoniDataCredential import *
from tuoni.TuoniJob import *
from tuoni.TuoniCommandPlugin import *
from tuoni.TuoniCommandTemplate import *


class TuoniC2:
    """
    The primary class for establishing connections to and managing interactions with the Tuoni server. It provides functionality for controlling server operations and facilitating communication.

    Args:
        verify (str | bool) [default = False]: A flag to enable or disable SSL verification. If a string is provided, it is treated as the path to a CA bundle file.
    """
    def __init__(self, verify: str | bool = False):
        self._token: str = None
        self._url: str = None
        self._monitoring_threads: list = []
        self._set_verify(verify)

    def login(self, url: str, username: str, password: str):
        """
        Login to the Tuoni server.

        Args:
            url (str): The URL of the Tuoni server to connect to.
            username (str): The username for authentication.
            password (str): The password for authentication.

        Examples:
            >>> tuoni_server = TuoniC2()
            # Disabling SSL verification
            >>> tuoni_server = TuoniC2(verify=False)
            # Set path for CA bunle
            >>> tuoni_server = TuoniC2(verify="/path/to/ca_bundle.pem")
            # Login to the server
            >>> tuoni_server.login("https://localhost:8443", "my_user", "S3cr37")


        """
        headers = {
            "Authorization": "Basic " + base64.b64encode(f"{username}:{password}".encode('utf-8')).decode('utf-8')
        }

        response = requests.post(f"{url}/api/v1/auth/login", headers=headers, verify=self._verify)
        if response.status_code < 200 or response.status_code >= 300:
            raise ExceptionTuoniAuthentication(response.text)
        self._token = response.text
        self._url = url

    def _request_check(self):
        if self._token is None:
            raise ExceptionTuoniAuthentication("You have not done the login")

    def _set_verify(self, verify: str | bool):
        self._verify = False
        if isinstance(verify, bool) and verify:
            self._verify = True
        elif isinstance(verify, bool) and not verify:
            self._verify = False
        elif isinstance(verify, str):
            if Path(verify).is_file():
                self._verify = verify
            else:
                raise Exception("The path to the CA bundle is not valid")


    @staticmethod
    def _raise_request_exception(result_msg: str):
        try:
            data = json.loads(result_msg)
            msg = data.get("message", result_msg)
        except json.JSONDecodeError:
            msg = result_msg
        raise ExceptionTuoniRequestFailed(msg)

    def _make_request(self, method: str, uri: str, **kwargs):
        self._request_check()
        headers = {"Authorization": f"Bearer {self._token}"}
        response = requests.request(method, f"{self._url}{uri}", headers=headers, verify=self._verify, **kwargs)
        if response.status_code < 200 or response.status_code >= 300:
            self._raise_request_exception(response.text)
        return response

    def request_get(self, uri: str, result_as_json: bool = True, result_as_bytes: bool = False):
        """
        Send a GET request to the Tuoni server.

        Args:
            uri (str): The URI endpoint to send the request to.
            result_as_json (bool): If True, the server's response is treated as JSON and converted to a dictionary.
            result_as_bytes (bool): If True, the server's response is returned as a bytes.

        Returns:
            str | dict | bytes: The server's response, either as a raw string or a dictionary if `result_as_json` is True or as bytes if `result_as_bytes` is True.
        """
        response = self._make_request("GET", uri)
        if len(response.content) == 0:
            return None
        if result_as_json:
            return json.loads(response.text)
        if result_as_bytes:
            return response.content
        return response.text

    def request_get_file(self, uri: str, file_name: str):
        """
        Send a GET request to the Tuoni server and save the result to the filesystem.

        Args:
            uri (str): The URI endpoint to send the request to.
            file_name (str): The name of the file where the response will be saved.
        """
        response = self._make_request("GET", uri, stream=True)
        with open(file_name, 'wb') as out_file:
            shutil.copyfileobj(response.raw, out_file)

    def request_post(self, uri: str, json_data: dict = None, files: dict = None):
        """
        Send a POST request to the Tuoni server.

        Args:
            uri (str): The URI endpoint to send the request to.
            json_data (dict): A dictionary containing the JSON payload to include in the POST request.
            files (dict): A dictionary of files to upload with the POST request.

        Returns:
            dict: The server's response as a dictionary.

        Examples:
            >>> tuoni = TuoniC2()
            >>> tuoni_server.request_post(
            >>>     "/api/v1/command-alias",
            >>>     {
            >>>         "name": "bofX",
            >>>         "description": "Example",
            >>>         "baseTemplate": "bof",
            >>>         "fixedConfiguration": {"method": "go"}
            >>>     },
            >>>     {"bofFile" : ["some_bof.o", open("some_bof", "rb").read()]}
            >>> )
        """
        if files is None:
            response = self._make_request("POST", uri, json=json_data)
        else:
            all_data = {}
            if json_data is not None:
                all_data["requestBody"] = (None, json.dumps(json_data), 'application/json')
            for var_name, file_info in files.items():
                all_data[var_name] = (file_info[0], file_info[1], 'application/octet-stream')
            response = self._make_request("POST", uri, files=all_data)
        return json.loads(response.text) if response.text else None

    def request_patch(self, uri: str, json_data: dict = None, files: dict = None):
        """
        Send a PATCH request to the Tuoni server.

        Args:
            uri (str): The URI endpoint to send the request to.
            json_data (dict): A dictionary containing the JSON payload to include in the POST request.
            files (dict): A dictionary of files to upload with the POST request.

        Returns:
            dict: The server's response as a dictionary.
        """
        if files is None:
            response = self._make_request("PATCH", uri, json=json_data)
        else:
            all_data = {}
            if json_data is not None:
                all_data["requestBody"] = (None, json.dumps(json_data), 'application/json')
            for var_name, file_info in files.items():
                all_data[var_name] = (file_info[0], file_info[1], 'application/octet-stream')
            response = self._make_request("PATCH", uri, files=all_data)
        return json.loads(response.text) if response.text else None

    def request_put(self, uri: str, json_data: dict = None):
        """
        Send a PUT request to the Tuoni server.

        Args:
            uri (str): The URI endpoint to send the request to.
            json_data (dict): A dictionary containing the JSON payload to include in the PUT request.

        Returns:
            dict: The server's response as a dictionary.
        """
        response = self._make_request("PUT", uri, json=json_data)
        return json.loads(response.text) if response.text else None

    def request_delete(self, uri: str, json_data: dict = None):
        """
        Send a DELETE request to the Tuoni server.

        Args:
            uri (str): The URI endpoint to send the request to.
            json_data (dict): A dictionary containing the JSON payload to include in the DELETE request.

        Returns:
            dict: The server's response as a dictionary.
        """
        response = self._make_request("DELETE", uri, json=json_data)
        return json.loads(response.text) if response.text else None

    def load_listener_plugins(self):
        """
        Retrieve a list of listener plugins.

        Returns:
            list[TuoniListenerPlugin]: A list of available listener plugins.

        Examples:
            >>> http_listener_plugin = tuoni_server.load_listener_plugins()[
            >>>     "shelldot.listener.agent-reverse-http"
            >>> ]
            >>> conf = http_listener_plugin.conf_examples["default"]
            >>> conf["sleep"] = 2
            >>> conf["instantResponses"] = True
            >>> listener = http_listener_plugin.create(conf)
        """
        plugins_data = self.request_get("/api/v1/plugins/listeners")
        return {plugin_data["identifier"]["id"]: TuoniListenerPlugin(plugin_data, self) for plugin_data in plugins_data.values()}

    def load_listeners(self):
        """
        Retrieve a list of listeners.

        Returns:
            list[TuoniListener]: A list of active listeners.
        """
        listeners_data = self.request_get("/api/v1/listeners")
        return [TuoniListener(listener_data, self) for listener_data in listeners_data.values()]

    def load_payload_plugins(self):
        """
        Retrieve a list of payload plugins.

        Returns:
            list[TuoniPayloadPlugin]: A list of available payload plugins.
        """
        plugins_data = self.request_get("/api/v1/plugins/payloads")
        return {plugin_data["identifier"]["id"]: TuoniPayloadPlugin(plugin_data, self) for plugin_data in plugins_data.values()}

    def create_payload(self, payload_template: str, payload_listener: int, payload_conf: dict, encrypted: bool = True, payload_name: str = None):
        """
        Create a new payload.

        Args:
            payload_template (str): The payload template to use.
            payload_listener (int): The ID of the listener associated with this payload.
            payload_conf (dict): A dictionary containing the payload's configuration.
            encrypted (bool): Specifies whether traffic for this payload should be encrypted.
            payload_name (str): The name to assign to the payload.

        Returns:
            id: The unique ID of the created payload.

        Examples:
            >>> payload_id = tuoni_server.create_payload(
            >>>     "shelldot.payload.windows-x64",
            >>>     listener_id,
            >>>     {"type": "executable"}
            >>> )
        """
        json_data = {
            "payloadTemplateId": payload_template,
            "configuration": payload_conf,
            "listenerId": payload_listener,
            "encrypted": encrypted,
            "name": payload_name
        }
        payload_data = self.request_post("/api/v1/payloads", json_data)
        return payload_data["id"]

    def download_payload(self, payload_id: int, file_name: str):
        """
        Download a payload.

        Args:
            payload_id (int): The unique ID of the payload to download.
            file_name (str): The name of the file to save the downloaded payload.
        """
        self.request_get_file(f"/api/v1/payloads/{payload_id}/download", file_name)

    def load_agents(self, active = True, unactive = False):
        """
        Retrieve a list of agents.
        
        Args:
            active (bool): Should active agents be returned
            unactive (bool): Should unactive agents be returned

        Returns:
            list[TuoniAgent]: A list of agents.
        """
        agents_data = []
        if active and not unactive:
            agents_data = self.request_get("/api/v1/agents/active")
        elif not active and unactive:
            agents_data = self.request_get("/api/v1/agents/inactive")
        elif active and unactive:
            agents_data = self.request_get("/api/v1/agents")
        return [TuoniAgent(agent_data, self) for agent_data in agents_data]

    def wait_new_agent(self, interval: int = 1, max_wait: int = 0):
        """
        Wait for a new agent to connect.

        Args:
            interval (int): The interval, in seconds, to check for new connections.
            max_wait (int): The maximum time to wait, in seconds. A value of 0 means to wait indefinitely.

        Returns:
            TuoniAgent: The newly connected agent.
        """
        original_agents = {agent_data["guid"] for agent_data in self.request_get("/api/v1/agents/active")}
        while True:
            time.sleep(interval)
            agents_data = self.request_get("/api/v1/agents/active")
            for agent_data in agents_data:
                if agent_data["guid"] not in original_agents:
                    return TuoniAgent(agent_data, self)
            if max_wait > 0:
                max_wait -= interval
                if max_wait <= 0:
                    return None

    def on_new_agent(self, function, interval: int = 1):
        """
        Set a callback function to be triggered when a new agent connects.

        Args:
            function (func): The function to execute when a new agent connects.
            interval (int): The interval, in seconds, to check for new connections.

        Examples:
            >>> def new_agent_callback(agent):
            >>>     print(
            >>>         f"We got outselves a new agent {agent.guid} from {agent.metadata['hostname']}"
            >>>     )
            >>>
            >>> tuoni_server.on_new_agent(new_agent_callback)
        """
        monitor_thread = threading.Thread(target=self._monitor_for_new_agents, args=(function, interval), daemon=True)
        monitor_thread.start()
        self._monitoring_threads.append(monitor_thread)

    def _monitor_for_new_agents(self, function, interval: int = 1):
        original_agents = {agent_data["guid"] for agent_data in self.request_get("/api/v1/agents")}
        while True:
            time.sleep(interval)
            agents_data = self.request_get("/api/v1/agents")
            for agent_data in agents_data:
                if agent_data["guid"] not in original_agents:
                    original_agents.add(agent_data["guid"])
                    agent = TuoniAgent(agent_data, self)
                    threading.Thread(target=function, args=(agent,)).start()

    def load_command_plugins(self):
        """
        Retrieve a list of command plugins.

        Returns:
            list[TuoniCommandPlugin]: A list of available command plugins.
        """
        plugins_data = self.request_get("/api/v1/plugins/commands")
        return {plugin_data["identifier"]["id"]: TuoniCommandPlugin(plugin_data, self) for plugin_data in plugins_data.values()}
    
    def load_command_templates(self):
        """
        Retrieve a list of command templates.

        Returns:
            list[TuoniCommandTemplate]: A list of available command templates.
        """
        templates_data = self.request_get("/api/v1/command-templates")
        return [TuoniCommandTemplate(template_data, self) for template_data in templates_data]
    
    def load_command_template(self, guid):
        """
        Retrieve a specific command template by its unique identifier.

        Args:
            guid (str): The unique identifier of the command template to retrieve.

        Returns:
            TuoniCommandTemplate: The requested command template, or None if not found.
        """
        template_data = self.request_get(f"/api/v1/command-templates/{guid}")
        if template_data:
            return TuoniCommandTemplate(template_data, self)
        return None

    def load_aliases(self):
        """
        Retrieve a list of aliases.

        Returns:
            list[TuoniAlias]: A list of aliases.
        """
        all_aliases = self.request_get("/api/v1/command-alias")
        return [TuoniAlias(alias_data, self) for alias_data in all_aliases]

    def add_alias(self, name, description, command_type, command_conf={}, files = None):
        """
        Add a new alias.

        Args:
            name (str): The name of the alias to create.
            description (str): A description of the alias.
            command_type (str | TuoniDefaultCommand): The base command to associate with the alias.
            command_conf (dict): Configuration settings for the alias.
            files (dict): File parameters associated with the alias.

        Returns:
            TuoniAlias: The newly created alias.

        Examples:
            >>> alias1 = tuoni_server.add_alias(
            >>>     "ls1",
            >>>     "Alias made with python lib based on 'ls' default command class",
            >>>     TuoniCommandLs,
            >>>     {"depth": 1}
            >>> )

            >>> alias2 = tuoni_server.add_alias(
            >>>     "ls2",
            >>>     "Alias made with python lib based on command string name",
            >>>     "ls",
            >>>     {"depth": 2}
            >>> )

            >>> alias3 = tuoni_server.add_alias(
            >>>     "easm",
            >>>     "Alias for execute assembly default command class",
            >>>     TuoniCommandexecuteAssembly,
            >>>     {},
            >>>     files = {
            >>>         "executable": ["dotnet.exe",open("dotnet.exe", "rb").read()]
            >>>     }
            >>> )

            >>> alias4 = tuoni_server.add_alias(
            >>>     "bof1",
            >>>     "Alias for bof based on command string name",
            >>>     "bof",
            >>>     files = {
            >>>         "bofFile": ["bof.o",open("bof.o", "rb").read()]
            >>>     }
            >>> )

            >>> alias5 = tuoni_server.add_alias(
            >>>     "bof2",
            >>>     "Alias for bof based on command ID",
            >>>     "2ac58f33-d35e-4afd-a1ac-00e460ceb9f4",
            >>>     files = {
            >>>         "bofFile": ["bof.o",open("bof.o", "rb").read()]
            >>>     }
            >>> )
        """
        if isinstance(command_type, TuoniDefaultCommand):
            command_conf = command_type.command_conf
            files = command_type.files
            command_type = command_type.command_type
        if inspect.isclass(command_type) and issubclass(command_type, TuoniDefaultCommand):
            command_type = command_type._class_base_type

        json_data = {
            "name": name,
            "description": description,
            "baseTemplate": command_type,
            "fixedConfiguration": command_conf
        }
        alias_data = self.request_post("/api/v1/command-alias", json_data, files = files)
        return TuoniAlias(alias_data, self)

    def load_hosted(self):
        """
        Retrieve a list of hosted files.

        Returns:
            list[TuoniFile]: A list of objects containing the details of hosted files.
        """
        all_files = self.request_get("/api/v1/files")
        return [TuoniFile(file_data, self) for file_data in all_files]

    def add_hosted(self, filename, file_content, original_filename = None):
        """
        Add a hosted file.

        Args:
            filename (str): The name of the file to host.
            file_content (bytes): The content of the file in bytes.

        Returns:
            str: The API URI for the uploaded file.

        Examples:
            >>> tuoni_server.add_hosted("/hosted/file/here.txt", b"HELLO WORLD")
        """
        if original_filename is None:
            original_filename = os.path.basename(filename)
        return self.request_post("/api/v1/files", files = {filename: [original_filename, file_content]})

    def delete_hosted(self, hosted):
        """
        Delete a hosted file.

        Returns:
            None
        """
        if isinstance(hosted, TuoniFile):
            self.request_delete("/api/v1/file/" + hosted.fileId)
        else:
            self.request_delete("/api/v1/file/" + hosted)

    def load_users(self):
        """
        Retrieve a list of users.

        Returns:
            list[TuoniUser]: A list of users.
        """
        all_users = self.request_get("/api/v1/users")
        return [TuoniUser(user_data, self) for user_data in all_users]

    def add_user(self, username, password, authorities):
        """
        Add a new user.

        Args:
            username (str): The username for the new user.
            password (str): The initial password for the user.
            authorities (list[str]): A list of authorities or roles assigned to the user.

        Returns:
            TuoniUser: The newly created user.

        Examples:
            >>> user = tuoni_server.add_user(
            >>>     "cool_new_user",
            >>>     "cool_new_password",
            >>>     [
            >>>         "MANAGE_LISTENERS",
            >>>         "MANAGE_USERS",
            >>>         "SEND_COMMANDS",
            >>>         "MANAGE_PAYLOADS",
            >>>         "MODIFY_FILES",
            >>>         "VIEW_RESOURCES",
            >>>         "MANAGE_AGENTS"
            >>>     ]
            >>> )
        """
        json_data = {
            "username": username,
            "password": password,
            "authorities": authorities
        }
        user_data = self.request_post("/api/v1/users", json_data, None)
        return TuoniUser(user_data, self)

    def load_datamodel_hosts(self, page = 0, pageSize = 256, filter = None):
        """
        Retrieve a list of hosts.

        Args:
            page (int): Page of the result.
            pageSize (int): How many results per page.
            filter (str): Additional filters.

        Returns:
            list[TuoniDataHost]: A list of hosts.
        """
        if filter is None:
            filter = ""
        else:
            filter = "&" + filter

        all_hosts = self.request_get(f"/api/v1/discovery/hosts?page={page}&pageSize={pageSize}{filter}")
        if "items" in all_hosts:
            return [TuoniDataHost(host_data, self) for host_data in all_hosts["items"]]
        return []

    def add_datamodel_host(self, address, name, note):
        """
        Add a host data model entry.

        Args:
            address (str): The host address.
            name (str): Given name to the host.
            note (str): Additional notes.

        Returns:
            TuoniDataHost: The newly created host.

        Examples:
            >>> host = tuoni_server.add_datamodel_host("10.20.30.40", "WS1", "Open windows machine")
        """
        json_data = {
            "address": address,
            "name": name,
            "note": note
        }
        host_data = self.request_post("/api/v1/discovery/hosts", json_data, None)
        return TuoniDataHost(host_data, self)

    def load_datamodel_services(self, page = 0, pageSize = 256, filter = None):
        """
        Retrieve a list of services.

        Args:
            page (int): Page of the result.
            pageSize (int): How many results per page.
            filter (str): Additional filters.

        Returns:
            list[TuoniDataService]: A list of services.
        """
        if filter is None:
            filter = ""
        else:
            filter = "&" + filter

        all_services = self.request_get(f"/api/v1/discovery/services?page={page}&pageSize={pageSize}{filter}")
        if "items" in all_services:
            return [TuoniDataService(service_data, self) for service_data in all_services["items"]]
        return []

    def add_datamodel_service(self, address, port, protocol, banner, note):
        """
        Add a service data model entry.

        Args:
            address (str): The service address.
            port (int): Port number.
            protocol (str): Service protocol.
            banner (str): Service banner.
            note (str): Additional notes.

        Returns:
            TuoniDataService: The newly created service.

        Examples:
            >>> service = tuoni_server.add_datamodel_service("10.20.30.40", "443", "HTTPS", "", "")
        """
        json_data = {
            "address": address,
            "port": port,
            "protocol": protocol,
            "banner": banner,
            "note": note
        }
        service_data = self.request_post("/api/v1/discovery/services", json_data, None)
        return TuoniDataService(service_data, self)

    def load_datamodel_credentials(self, page = 0, pageSize = 256, filter = None):
        """
        Retrieve a list of credentials.

        Args:
            page (int): Page of the result.
            pageSize (int): How many results per page.
            filter (str): Additional filters.

        Returns:
            list[TuoniDataCredential]: A list of credentials.
        """
        if filter is None:
            filter = ""
        else:
            filter = "&" + filter

        all_credentials = self.request_get(f"/api/v1/discovery/credentials?page={page}&pageSize={pageSize}{filter}")
        if "items" in all_credentials:
            return [TuoniDataCredential(credential_data, self) for credential_data in all_credentials["items"]]
        return []

    def add_datamodel_credential(self, username, password, host, realm, source, note):
        """
        Add a credential data model entry.

        Args:
            username (str): The username.
            password (str): The password.
            host (str): Host where credential works.
            realm (str): Credential realm.
            source (str): Source of the credential.
            note (str): Additional notes.

        Returns:
            TuoniDataCredential: The newly created credential.

        Examples:
            >>> credential = tuoni_server.add_datamodel_credential(
            >>>     "rick",
            >>>     "NeverGonnaBypassYourEDR",
            >>>     "10.20.30.40",
            >>>     "", "", ""
            >>> )
        """
        json_data = {
            "username": username,
            "password": password,
            "host": host,
            "realm": realm,
            "source": source,
            "note": note
        }
        credential_data = self.request_post("/api/v1/discovery/credentials", json_data, None)
        return TuoniDataCredential(credential_data, self)

    def load_jobs(self, page = 0, pageSize = 256, sort_col = "id", sort_order = "asc", inactives = False):
        """
        Retrieve a list of jobs.

        Args:
            page (int): Page of the result.
            pageSize (int): How many results per page.
            sort_col (str): Based of what field to sort
            sort_order (str): Order of sort
            inactives (bool): Should also inactive jobs be returned


        Returns:
            list[TuoniJob]: A list of jobs.
        """
        sort_col = "&" + sort_col + ":" + sort_order
        urlEnd = "all" if inactives else "active"
        all_jobs = self.request_get(f"/api/v1/jobs/{urlEnd}?page={page}&pageSize={pageSize}{sort_col}")
        if "items" in all_jobs:
            return [TuoniJob(job_data, self) for job_data in all_jobs["items"]]
        return []

    def let_it_run(self):
        """
        Block execution and wait indefinitely if any callback functions are initialized,
        or until all monitoring threads have completed.
        """
        for monitoring_thread in self._monitoring_threads:
            monitoring_thread.join()



import random
from tuoni.TuoniCommandTemplate import *

class TuoniCommandPlugin:
    """
    A class that provides data and functionality for a command plugin.

    Attributes:
        name (str): The name of the command plugin.
        vendor (str): The vendor of the command plugin.
        description (str): A description of the command plugin.
        plugin_id (str): The unique identifier of the command plugin.
        commands (list[TuoniCommandTemplate]): A list of command templates associated with the plugin.
        
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the command plugin class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self.name = conf["info"]["name"]
        self.vendor = conf["info"]["vendor"]
        self.description = conf["info"]["description"]
        self.plugin_id = conf["identifier"]["id"]
        self.c2 = c2
        self.commands = []
        for command_name in conf["commands"]:
            self.commands.append(TuoniCommandTemplate(conf["commands"][command_name], c2))



from tuoni.TuoniExceptions import *
from tuoni.TuoniCommand import *
from tuoni.TuoniAlias import *
from tuoni.TuoniDefaultCommands import *
from tuoni.TuoniCommandTemplate import *
import warnings

class TuoniAgent:
    """
    A class that encapsulates the data and functionality of a connected agent.

    Attributes:
        guid (GUID): A unique identifier (GUID) assigned to the agent.
        first_registration_time (str): The time, in string format, when the agent first connected.
        last_callback_time (str): The time, in string format, of the agent's most recent connection.
        metadata (dict): A dictionary containing metadata about the agent.
        active (bool): A boolean flag indicating whether the agent is currently active.
        recentListeners (list): A list of listeners that the agent uses to maintain its connection.
        availableCommands (dict): A dictionary of commands that the agent is capable of executing.
    """

    def __init__(self, conf, c2):
        """
        Constructor.

        Attributes:
            conf (dict): Data from server.
            c2 (TuoniC2): Related server object.
        
        """
        self.c2 = c2
        self._load_conf(conf)

    def _load_conf(self, conf):
        self.guid = conf["guid"]
        self.first_registration_time = conf["firstRegistrationTime"]
        self.last_callback_time = conf["lastCallbackTime"]
        self.metadata = conf["metadata"]
        self.active = conf["active"]
        self.recentListeners = conf["recentListeners"]
        self._fill_available_commands(conf["availableCommandTemplates"])

    def send_command(self, command_type, command_conf=None, execution_conf = None, files = None):
        """
        Send command to agent.

        Args:
            command_type (str | TuoniAlias | TuoniDefaultCommand | TuoniCommandTemplate): What command to send.
            command_conf (dict): Command configuration.
            execution_conf (dict): Execution configuration.
            files (dict): Files to send with command.

        Returns:
            TuoniCommand: An object representing the created command.

        Examples:
            >>> command1 = agent.send_command(TuoniCommandLs(".\\subdir", 2))
            >>> command2 = agent.send_command(
            >>>     TuoniCommandProcinfo(), 
            >>>     execution_conf ={"execType": "NEW"}
            >>> )
            >>> command3 = agent.send_command(
            >>>     TuoniCommandProcinfo(), 
            >>>     execution_conf = {
            >>>         "execType": "EXISTING", 
            >>>         "pid": 1234
            >>>     }
            >>> )
            >>> command4 = agent.send_command("ls", {"dir": ".\\subdir", "depth": 2})
            >>> command5 = agent.send_command(
            >>>     alias_object, 
            >>>     {"conf_value_name": "conf_value"}
            >>> )
        """
        if self.guid is None:
            raise ExceptionTuoniDeleted("")
        if isinstance(command_type, TuoniDefaultCommand):
            command_conf = command_type.command_conf
            execution_conf = command_type.execution_conf
            files = command_type.files
            command_type = command_type.command_type
        if isinstance(command_type, TuoniCommandTemplate):
            command_type = command_type.id
        if isinstance(command_type, TuoniAlias):
            command_type = command_type.alias_id
        if command_conf is None:
            command_conf = {}
        data = {
            "template": command_type,
            "configuration": command_conf
        }
        if execution_conf is not None:
            data["execConf"] = execution_conf
        data = self.c2.request_post(f"/api/v1/agents/{self.guid}/commands", data, files)
        return TuoniCommand(data, self.c2)

    def get_commands(self):
        """
        Retrieves a list of all commands associated with the agent.

        Returns:
            list[TuoniCommand]: List of commands sent to the agent.
        """
        if self.guid is None:
            raise ExceptionTuoniDeleted("")
        commands_data = self.c2.request_get(f"/api/v1/agents/{self.guid}/commands")
        commands = []
        for command_nr in commands_data:
            command_data = commands_data[command_nr]
            command_obj = TuoniCommand(command_data, self.c2)
            commands.append(command_obj)
        return commands

    def delete(self):
        """
        Deletes agent (actually makes it inactive but close enough).
        """
        if self.guid is None:
            raise ExceptionTuoniDeleted("")
        self.c2.request_put(f"/api/v1/agents/{self.guid}/inactive")
        self.listener_id = None
        
    def _fill_available_commands(self, command_list):
        self.availableCommands = {}
        for cmd in self.c2.request_get("/api/v1/command-templates"):
            if cmd["id"] in command_list:
                self.availableCommands[cmd["name"]] = cmd["id"]

    def matches(self, criteria):
        """
        Check if the agent matches the given criteria.

        Args:
            criteria (dict): Dictionary containing matching criteria. Keys can 
            use dot notation like 'metadata.os' to access nested attributes.

        Returns:
            bool: True if all criteria match, False otherwise.
        """
        
        for key, expected_value in criteria.items():
            # Handle dot notation for nested attributes
            current_obj = self
            
            # Split the key by dots and traverse the object
            key_parts = key.split('.')
            
            try:
                for part in key_parts:
                    if hasattr(current_obj, part):
                        current_obj = getattr(current_obj, part)
                    elif isinstance(current_obj, dict) and part in current_obj:
                        current_obj = current_obj[part]
                    else:
                        # Debug: print available attributes
                        if hasattr(current_obj, '__dict__'):
                            available_attrs = list(current_obj.__dict__.keys())
                        elif isinstance(current_obj, dict):
                            available_attrs = list(current_obj.keys())
                        else:
                            available_attrs = []
                        return False
                
                # Compare the final value (case-insensitive for strings)
                if isinstance(current_obj, str) and isinstance(expected_value, str):
                    if current_obj.upper() != expected_value.upper():
                        return False
                else:
                    if current_obj != expected_value:
                        print(f"{Fore.YELLOW}Debug: Value mismatch for '{key}': got '{current_obj}', expected '{expected_value}'{Style.RESET_ALL}")
                        return False
                        
            except (AttributeError, TypeError) as e:
                print(f"{Fore.YELLOW}Debug: Exception accessing '{key}': {e}{Style.RESET_ALL}")
                return False
        
        # All criteria matched
        return True
    
    def setCustomProperties(self, name, value):
        """
        .. deprecated
            Use :func:`set_custom_property` instead.

        Adds or updates a property in the agents customProperties metadata.
        

        Parameters:
            name (str): The name (key) of the property to add or update.
            value (Any): The value to assign to the given name.

        Behavior:
            - If the property with the given name exists, its value will be updated.
            - If it does not exist, a new name/value pair will be added.
            - If value is None the name/value pair will be deleted.

        Example:
            myAgent.setCustomProperties("notes", "Domain Controller")
        """
        warnings.warn(
            f"Function setCustomProperties is deprecated; it will be removed in a future release. Please use `set_custom_property` instead.",
            category=DeprecationWarning
        )
        self.set_custom_property(name, value)

    def set_custom_property(self, name, value):
        """
        Adds or updates a property in the agents customProperties metadata.

        Parameters:
            name (str): The name (key) of the property to add or update.
            value (Any): The value to assign to the given name.

        Behavior:
            - If the property with the given name exists, its value will be updated.
            - If it does not exist, a new name/value pair will be added.
            - If value is None the name/value pair will be deleted.

        Example:
            myAgent.set_custom_property("notes", "Domain Controller")
        """
        self.metadata['customProperties'][name] = value
        self.c2.request_put(f"/api/v1/agents/{self.guid}/metadata", self.metadata)



import time
from tuoni.TuoniExceptions import *


class TuoniUser:
    """
    A class that provides data and functionality for users.

    Attributes:
        username (str): The username of the user.
        enabled (bool): Indicates whether the user is enabled.
        authorities (list[str]): A list of authorities or roles assigned to the user.
    """
    def __init__(self, conf, c2):
        """
        Constructor for the result part class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.username = conf["username"]
        self.enabled = conf["enabled"]
        self.authorities = conf["authorities"]

    def reload(self):
        """
        Reload the user data from the C2 server.
        """
        if self.alias_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/users/{self.username}")
        self._load_conf(data)

    def disable(self):
        """
        Disable the user.
        """
        data = {"enabled": False, "authorities": self.authorities}
        data = self.c2.request_put(f"/api/v1/users/{self.username}", data)
        self._load_conf(data)

    def enable(self):
        """
        Enable the user.
        """
        data = {"enabled": True, "authorities": self.authorities}
        data = self.c2.request_put(f"/api/v1/users/{self.username}", data)
        self._load_conf(data)

    def set_authorities(self, authorities):
        """
        Change the authorities assigned to the user.
        """
        data = {"authorities": authorities, "enabled": self.enabled}
        data = self.c2.request_put(f"/api/v1/users/{self.username}", data)
        self._load_conf(data)

    def set_password(self, new_password):
        """
        Set a new password for the user.
        """
        data = {"newPassword": new_password}
        data = self.c2.request_put(f"/api/v1/users/{self.username}/password", data)
        
    




import random
from tuoni.TuoniListener import *

class TuoniListenerPlugin:
    """
    A class that provides data and functionality for a listener plugin.

    Attributes:
        name (str): The name of the listener plugin.
        vendor (str): The vendor of the listener plugin.
        description (str): A description of the listener plugin.
        plugin_id (str): The unique identifier of the listener plugin.
        conf_schema (dict): The configuration schema for the listener plugin.
        conf_examples (dict): Examples of valid configurations for the listener plugin.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the listener plugin class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self.name = conf["info"]["name"]
        self.vendor = conf["info"]["vendor"]
        self.description = conf["info"]["description"]
        self.plugin_id = conf["identifier"]["id"]
        self.conf_schema = conf["configurationSchema"]
        self.conf_examples = {}
        if "defaultConfiguration" in conf:
            self.conf_examples["default"] = conf["defaultConfiguration"]
        if "exampleConfigurations" in conf:
            for example in conf["exampleConfigurations"]:
                self.conf_examples[example["name"]] = example["configuration"]
        self.c2 = c2

    def create(self, new_listener_conf, new_listener_name=None, new_listener_conf_files=None):
        """
        Create a new listener.

        Args:
            new_listener_conf (dict): The configuration for the new listener.
            new_listener_name (str): The name to assign to the new listener.
            new_listener_conf_files (list): A list of tuples containing parameter names, file names, and file contents.
            
        Returns:
            TuoniListener: An object representing the newly created listener.

        Examples:
            >>> http_listener_plugin = tuoni_server.load_listener_plugins()[
            >>>     "shelldot.listener.agent-reverse-http"
            >>> ]
            >>> conf = http_listener_plugin.conf_examples["default"]
            >>> conf["sleep"] = 2
            >>> conf["instantResponses"] = True
            >>> listener = http_listener_plugin.create(conf)
        """
        json_data = {
            "plugin": self.plugin_id,
            "configuration": new_listener_conf
        }
        if new_listener_name is not None:
            json_data["name"] = new_listener_name
        files_data = None
        if new_listener_conf_files:
            files_data = {}
            for param_name, file_name, file_content in new_listener_conf_files:
                files_data[param_name] = (file_name, file_content)
        listener_data = self.c2.request_post("/api/v1/listeners", json_data, files=files_data)
        listener_obj = TuoniListener(listener_data, self.c2)
        return listener_obj

    def get_default_conf(self):    
        """
        Retrieve the default configuration for the listener plugin.

        Returns:
            dict: The default configuration settings, or an empty dictionary if none are defined.
        """
        if "default" in self.conf_examples:
            return self.conf_examples["default"]
        return {}  #Might change but let's say for now that if no "default" conf then empty conf is same

    def get_minimal_conf(self):
        """
        Retrieve the minimal configuration for the listener plugin.

        Returns:
            dict: The minimal configuration settings, or an empty dictionary if none are defined.
        """
        return self.get_default_conf()




import json
from tuoni.TuoniExceptions import *


class TuoniResultPart:
    """
    A class that provides data and functionality for a part of a command result.

    Attributes:
        type (str): The type of the result part.
        name (str): The name of the result part.
        value (str): The value associated with the result part.
        filename (str): The filename of any file included in the result part.
        uri (str): The URI for accessing the file in the result part.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the result part class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.type = conf["type"]
        if self.type == "text":
            self.name = conf["name"]
            self.value = conf["value"]
        if self.type == "file":
            self.filename = conf["filename"]
            self.href = conf["href"]
            if "//" in self.href:
                self.uri = self.href[self.href.find("//")+2:]
                self.uri = self.uri[self.uri.find("/"):]
            else:
                self.uri = self.href

    def get_as_text(self):
        """
        Retrieve the value of the result part as text.

        Returns:
            str: The value of the result part as a string.
        """
        if self.type == "text":
            return self.value
        if self.type == "file":
            return self.c2.request_get(self.uri, result_as_json=False)
        return None

    def get_as_file(self, filename):
        """
        Retrieve the value of the result part as a file and save it to the specified location.

        Args:
            filename (str): The path where the file should be saved.
        """
        if self.type == "file":
            self.c2.request_get_file(self.uri, filename)
        return None

    def get_as_bytes(self):
        """
        Retrieve the value of the result part as bytes.
        """        
        if self.type == "file":
            return self.c2.request_get(self.uri, result_as_json=False, result_as_bytes=True)
        if self.type == "text":
            return self.value.encode('utf-8')
        return None

    def get_as_json(self):
        """
        Retrieve the value of the result part as JSON and convert it to a dictionary.

        Returns:
            dict: The result part as a dictionary.
        """
        if self.type == "text":
            return json.loads(self.value)
        if self.type == "file":
            return json.loads(self.c2.request_get(self.uri, result_as_json=False))
        return None



import time
from tuoni.TuoniEvent import *
from tuoni.TuoniExceptions import ExceptionTuoniDeleted


class TuoniJob:
    """
    A class that provides data and functionality for a job.

    Attributes:
        job_id (int): The unique identifier of the job.
        name (str): Name of the job.
        status (str): Status of the job.
        source (dict): Source for the job.
        supportedActions (list): What actions are allowed on the job.
        openResources (list): What resources are open related to this job.
        messages (list): Messages related to job.
        createEvent (TuoniEvent): Event about job creation.
        lastUpdateEvent (TuoniEvent): Event about last job update.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the job class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.job_id = conf["id"]
        self.name = conf["name"]
        self.status = conf["status"]
        self.source = conf["source"]
        self.supportedActions = conf["supportedActions"]
        self.openResources = conf["openResources"]
        self.messages = conf["messages"]
        self.createEvent = TuoniEvent(conf["createEvent"]) if (conf["createEvent"] is not None) else None
        self.lastUpdateEvent = TuoniEvent(conf["lastUpdateEvent"]) if (conf["lastUpdateEvent"] is not None) else None

    def reload(self):
        """
        Reload the command data from the C2 server.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/job/{self.job_id}")
        self._load_conf(data)

    def restart(self):
        """
        Run restart operation on the job.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/job/{self.job_id}/restart")

    def pause(self):
        """
        Run pause operation on the job.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/job/{self.job_id}/pause")

    def resume(self):
        """
        Run resume operation on the job.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/job/{self.job_id}/resume")


class ExceptionTuoniAuthentication(Exception):
    """
    Exception class raised for authentication-related issues.
    """
    
    def __init__(self, message=None):
        """
        Constructor for the authentication exception.

        Args:
            message (str): The error message describing the authentication issue.
        """
        self.message = message
        super().__init__(message)

class ExceptionTuoniRequestFailed(Exception):
    """
    Exception class raised when an HTTP request fails.
    """
    
    def __init__(self, message=None):
        """
        Constructor for the HTTP request failure exception.

        Args:
            message (str): The error message describing the reason for the failed HTTP request.
        """
        self.message = message
        super().__init__(message)

class ExceptionTuoniDeleted(Exception):
    """
    Exception class raised when attempting to interact with a deleted resource.
    """
    
    def __init__(self, message=None):
        """
        Constructor for the deleted resource exception.

        Args:
            message (str): The error message describing the interaction attempt with the deleted resource.
        """
        self.message = message
        super().__init__(message)


from tuoni.TuoniResultPart import *


class TuoniResult:
    """
    A class that provides data for a command result.

    Attributes:
        status (str): The status of the result.
        error_message (str): The error message, if any, for the result.
        received (str): The time, in string format, indicating when the result was received.
        parts (list[TuoniResultPart]): A list of parts that make up the result.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the result class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self.c2 = c2
        self._load_conf(conf)

    def _load_conf(self, conf):
        self.status = conf["status"]
        self.error_message = conf["errorMessage"]
        self.received = conf["received"]
        self.parts = []
        for part_conf in conf["childResults"]:
            part_obj = TuoniResultPart(part_conf, self.c2)
            self.parts.append(part_obj)

    def is_text(self):
        """
        Check if the result contains typical text output (result part where name is "stdout", "text", or "message").

        Returns:
            bool: True if the result is a text output, False otherwise.
        """
        for part in self.parts:
            if part.type == "text" and (part.name.lower() == "stdout" or part.name.lower() == "text" or part.name.lower() == "message"):
                return True
        return False
    
    def is_json(self):
        """
        Check if the result contains JSON output (result part where name "json").

        Returns:
            bool: True if the result is a JSON output, False otherwise.
        """
        for part in self.parts:
            if part.type == "text" and part.name.lower() == "json":
                return True
        return False
    
    def is_file(self):
        """
        Check if the result contains file output (result part where type is "file").

        Returns:
            bool: True if the result is a file output, False otherwise.
        """
        for part in self.parts:
            if part.type == "file":
                return True
        return False
    
    def get_text(self):
        """
        Retrieve the text output from the result parts.

        Returns:
            str: The text output if available, otherwise None.
        """
        for part in self.parts:
            if part.type == "text" and (part.name.lower() == "stdout" or part.name.lower() == "text" or part.name.lower() == "message"):
                return part.get_as_text()
        return None
    
    def get_json(self):
        """
        Retrieve the JSON output from the result parts.

        Returns:
            dict: The JSON output if available, otherwise None.
        """
        for part in self.parts:
            if part.type == "text" and part.name.lower() == "json":
                return part.get_as_json()
        return None
    
    def get_files(self, download = False):
        """
        Retrieve the file outputs from the result parts.

        Returns:
            dict: A dictionary where keys are filenames and values are the file contents if download is True, otherwise a ResultPart object.
        """
        files = {}
        for part in self.parts:
            if part.type == "file":
                if download:
                    files[part.filename] = part.get_as_bytes()
                else:
                    files[part.filename] = part
        return files




import time
from tuoni.TuoniExceptions import *
from tuoni.TuoniResult import *


class TuoniCommand:
    """
    A class that provides data and functionality for a sent command.

    Attributes:
        command_id (int): The unique identifier of the command.
        commandTemplateId (str): The unique identifier of the command template used to create the command.
        commandTemplateName (str): The name of the command template used to create the command.
        configuration (dict): The configuration settings for the command.
        execConf (dict): The execution configuration for the command.
        created (str): The time, as a string, indicating when the command was created.
        sent (str): The time, as a string, indicating when the command was sent.
        result (TuoniResult): The result data returned from the command execution.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the command class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.command_id = conf["id"]
        self.commandTemplateId = conf["commandTemplateId"]
        self.commandTemplateName = conf["commandTemplateName"]
        self.configuration = conf["configuration"]
        self.execConf = conf["execConf"]
        self.created = conf["created"]
        self.sent = conf["sent"]
        self.result = conf["result"]

    def reload(self):
        """
        Reload the command data from the C2 server.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/commands/{self.command_id}")
        self._load_conf(data)

    def result_status_done(self, reload = True):
        """
        Check if the command has completed.
    
        Args:
            reload (bool): If True, reload the command data from the server to verify the current status.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        self.reload()
        if self.result is None:
            return False
        if self.result["status"].lower() == "ongoing":
            return False
        return True

    def result_status_ongoing(self, reload = True):
        """
        Check if the command is still ongoing.

        Args:
            reload (bool): If True, reload the command data from the server to verify the current status.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        self.reload()
        if self.result is None:
            return False
        if self.result["status"].lower() == "ongoing":
            return True
        return False

    def get_result(self, reload = True):
        """
        Retrieve the result object of the command.

        Args:
            reload (bool): If True, reload the command result data from the server.

        Returns:
            TuoniResult: The result object containing the data from the command execution.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        self.reload()
        if self.result is None:
            return None
        return TuoniResult(self.result, self.c2)

    def wait_result(self, interval = 1, maxWait = 0):
        """
        Wait for the command result object to be available (command done or in case of ongoing result, the initial results)

        Args:
            interval (int): The interval, in seconds, to reload the data from the server.
            maxWait (int): The maximum time to wait, in seconds, before giving up. A value of 0 means wait indefinitely.

        Returns:
            TuoniResult: The result object containing the data from the command execution, if available within the specified time.
        """
        self.reload()
        while self.result is None:
            time.sleep(interval)
            self.reload()
            if maxWait > 0:
                maxWait -= interval
                if maxWait <= 0:
                    return None
        return TuoniResult(self.result, self.c2)

    def wait_done(self, interval = 1, maxWait = 0):
        """
        Wait for the command to be done (success or failed).

        Args:
            interval (int): The interval, in seconds, to reload the data from the server.
            maxWait (int): The maximum time to wait, in seconds, before giving up. A value of 0 means wait indefinitely.

        Returns:
            TuoniResult: The result object containing the data from the command execution, if available within the specified time.
        """
        self.reload()
        while self.result is None or self.result["status"].lower() == "ongoing":
            time.sleep(interval)
            self.reload()
            if maxWait > 0:
                maxWait -= interval
                if maxWait <= 0:
                    return None
        return TuoniResult(self.result, self.c2)
    
    def wait_sent(self, interval = 1, maxWait = 0):
        """
        Wait for the command to be sent

        Args:
            interval (int): The interval, in seconds, to reload the data from the server.
            maxWait (int): The maximum time to wait, in seconds, before giving up. A value of 0 means wait indefinitely.

        Returns:
            bool: Was command sent
        """
        self.reload()
        while self.sent is None:
            time.sleep(interval)
            self.reload()
            if maxWait > 0:
                maxWait -= interval
                if maxWait <= 0:
                    return False
        return True

    def is_done(self, reload_from_server = False):
        """
        Checks has the command done/finished

        Args:
            reload_from_server (bool): Should command data be reloaded from server.

        Returns:
            bool: Is command done
        """
        if reload_from_server:
            self.reload()
        return (self.result is not None and self.result["status"].lower() != "ongoing")

    def is_success(self, reload_from_server = False):
        """
        Checks was the command successful

        Args:
            reload_from_server (bool): Should command data be reloaded from server.

        Returns:
            bool: Was command successful
        """
        if reload_from_server:
            self.reload()
        return (self.result is not None and self.result["status"].lower() == "success")

    def is_failed(self, reload_from_server = False):
        """
        Checks did the command fail

        Args:
            reload_from_server (bool): Should command data be reloaded from server.

        Returns:
            bool: Did command fail
        """
        if reload_from_server:
            self.reload()
        return (self.result is not None and self.result["status"].lower() == "failed")

    def is_ongoing(self, reload_from_server = False):
        """
        Checks is the command running as ongoing command

        Args:
            reload_from_server (bool): Should command data be reloaded from server.

        Returns:
            bool: Is command ongoing
        """
        if reload_from_server:
            self.reload()
        return (self.result is not None and self.result["status"].lower() == "ongoing")

    def is_sent(self, reload_from_server = False):
        """
        Checks was the command sent to the agent

        Args:
            reload_from_server (bool): Should command data be reloaded from server.

        Returns:
            bool: Is command sent to agent
        """
        if reload_from_server:
            self.reload()
        return (self.sent is not None)
    
    def get_template(self):
        """
        Retrieve the command template used to create this command.

        Returns:
            TuoniCommandTemplate: The command template object, or None if the template could not be found.
        """
        return self.c2.load_command_template(self.commandTemplateId)



"""
tuoni: A Python library for using tuoni.io

This package provides tools to automate usage of Tuoni server, providing same functionality as GUI


"""
import inspect

from .TuoniAgent import TuoniAgent
from .TuoniAlias import TuoniAlias
from .TuoniC2 import TuoniC2
from .TuoniCommand import TuoniCommand
from .TuoniDefaultCommands import *
from .TuoniExceptions import *
from .TuoniListener import TuoniListener
from .TuoniListenerPlugin import TuoniListenerPlugin
from .TuoniPayloadPlugin import TuoniPayloadPlugin
from .TuoniResult import TuoniResult
from .TuoniResultPart import TuoniResultPart
from .TuoniUser import TuoniUser
from .TuoniFile import TuoniFile
from .TuoniDataHost import TuoniDataHost
from .TuoniDataService import TuoniDataService
from .TuoniDataCredential import TuoniDataCredential
from .TuoniJob import TuoniJob
from .TuoniEvent import TuoniEvent
from .TuoniCommandPlugin import TuoniCommandPlugin
from .TuoniCommandTemplate import TuoniCommandTemplate


__all__ = [
    'TuoniAgent',
    'TuoniAlias',
    'TuoniC2',
    'TuoniCommand',
    'TuoniListener',
    'TuoniListenerPlugin',
    'TuoniPayloadPlugin',
    'TuoniResult',
    'TuoniResultPart',
    'TuoniUser',
    'TuoniFile',
    "TuoniDataHost",
    "TuoniDataService",
    "TuoniDataCredential",
    "TuoniJob",
    "TuoniEvent",
    "TuoniCommandPlugin",
    "TuoniCommandTemplate",
    
    'ExceptionTuoniAuthentication',
    'ExceptionTuoniRequestFailed',
    'ExceptionTuoniDeleted',
    
    'ExecutionNew',
    'ExecutionExisting',
    'TuoniCommandBof',
    'TuoniCommandCd',
    'TuoniCommandDie',
    'TuoniCommandLs',
    'TuoniCommandCmd',
    'TuoniCommandJobs',
    'TuoniCommandProclist',
    'TuoniCommandRun',
    'TuoniCommandPowershell',
    'TuoniCommandSleep',
    'TuoniCommandTokenSteal',
    'TuoniCommandTokenDeleteAll',
    'TuoniCommandTokenDelete',
    'TuoniCommandTokenList',
    'TuoniCommandTokenMake',
    'TuoniCommandTokenUse',
    'TuoniCommandRm',
    'TuoniCommandDownload',
    'TuoniCommandUpload',
    'TuoniCommandCp',
    'TuoniCommandMv',
    'TuoniCommandMkdir',
    'TuoniCommandSocks5',
    'TuoniCommandConnectTcp',
    'TuoniCommandexecuteAssembly',
    'TuoniCommandInject',
    'TuoniCommandProcinfo',
    'TuoniCommandScreenshot',
    'TuoniCommandSpawn',
    'TuoniCommandJumpService',
    'TuoniCommandJumpWinrm',
    'TuoniCommandJumpWmi',
    'TuoniCommandJumpSsh',
    'TuoniCommandReverseShellCommunication'
]


import time
from tuoni.TuoniExceptions import *


class TuoniFile:
    """
    A class that provides data and functionality for a stored files.

    Attributes:
        fileId (GUID): The unique identifier (GUID) for the file.
        originalFileName (str): The original filename.
        size (int): Size of the file.
        filePaths (list): List of filepaths
        downloadHref (str): A download path.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the alias class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.fileId = conf["fileId"]
        self.originalFileName = conf["originalFileName"]
        self.size = conf["size"]
        self.filePaths = conf["filePaths"]
        self.downloadHref = conf["downloadHref"]

    def reload(self):
        """
        Reload the file data from the C2 server.
        """
        if self.fileId is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/file/{self.fileId}")
        self._load_conf(data)

    def delete(self):
        """
        Delete the file.
        """
        if self.fileId is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_delete(f"/api/v1/file/{self.fileId}")
        self._load_conf(data)

    def download(self, filename):
        return self.c2.request_get_file(self.downloadHref, filename)




import random
from tuoni.TuoniListener import *

class TuoniPayloadPlugin:
    """
    A class that provides data for a payload plugin.

    Attributes:
        name (str): The name of the payload plugin.
        vendor (str): The vendor of the payload plugin.
        description (str): A description of the payload plugin.
        plugin_id (str): The unique identifier of the payload plugin.
        templates (list): A list of available payload templates.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the payload plugin class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self.name = conf["info"]["name"]
        self.vendor = conf["info"]["vendor"]
        self.description = conf["info"]["description"]
        self.plugin_id = conf["identifier"]["id"]
        self.templates = []
        for payloadTemplate in conf["payloads"]:
            self.templates.append(payloadTemplate)
        self.c2 = c2




import random
from tuoni.TuoniListener import *

class TuoniCommandTemplate:
    """
    A class that provides data and functionality for a command template.

    Attributes:
        id (str): The unique identifier of the command template.
        name (str): The name of the command template.
        plugin_id (str): The unique identifier of the command plugin.
        scope (str): The scope of the command template.
        qualifiedName (str): The qualified name of the command template.
        fullyQualifiedName (str): The fully qualified name of the command template.
        description (str): A description of the command template.
        conf_schema (dict): The configuration schema for the command template.
        conf_examples (dict): Examples of valid configurations for the command template.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the command template class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self.id = conf["id"]
        self.name = conf["name"]
        self.plugin_id = conf["pluginId"]
        self.scope = conf["scope"]
        self.qualifiedName = conf["qualifiedName"]
        self.fullyQualifiedName = conf["fullyQualifiedName"]
        self.description = conf["description"]
        self.conf_schema = conf["configurationSchema"]
        self.conf_examples = {}
        if "defaultConfiguration" in conf:
            self.conf_examples["default"] = conf["defaultConfiguration"]
        if "exampleConfigurations" in conf:
            for example in conf["exampleConfigurations"]:
                self.conf_examples[example["name"]] = example["configuration"]
        self.c2 = c2

    def get_default_conf(self):    
        """
        Retrieve the default configuration for the command template.

        Returns:
            dict: The default configuration settings, or an empty dictionary if none are defined.
        """
        if "default" in self.conf_examples:
            return self.conf_examples["default"]
        return {}  #Might change but let's say for now that if no "default" conf then empty conf is same

    def get_minimal_conf(self):
        """
        Retrieve the minimal configuration for the command template.

        Returns:
            dict: The minimal configuration settings, or an empty dictionary if none are defined.
        """
        return self.get_default_conf()




import time
from tuoni.TuoniExceptions import ExceptionTuoniDeleted


class TuoniEvent:
    """
    A class that provides data and functionality for a event.

    Attributes:
        event_id (GUID): The unique identifier of the event.
        eventType (string): Type of the event.
        time (*): Event type.
        actor (dict): Actor that event is about.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the event class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.event_id = conf["id"]
        self.eventType = conf["eventType"]
        self.time = conf["time"]
        self.actor = conf["actor"]

    def reload(self):
        """
        Reload the command data from the C2 server.
        """
        if self.command_id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/event/{self.event_id}")
        self._load_conf(data)


import time
from tuoni.TuoniExceptions import *


class TuoniDataHost:
    """
    A class that provides data and functionality for a host data model entry.

    Attributes:
        id (GUID): The unique identifier (GUID) for the host.
        address (str): The host address.
        name (str): Given name to the host.
        note (str): Additional notes.
        status (str): Status of the entry.
    """
    
    def __init__(self, conf, c2):
        """
        Constructor for the host data class.

        Args:
            conf (dict): Data from the server.
            c2 (TuoniC2): The related server object that manages communication.
        """
        self._load_conf(conf)
        self.c2 = c2

    def _load_conf(self, conf):
        self.id = conf["id"]
        self.address = conf["address"]
        self.name = conf["name"]
        self.note = conf["note"]
        self.status = conf["status"]

    def reload(self):
        """
        Reload the host data from the C2 server.
        """
        if self.id is None:
            raise ExceptionTuoniDeleted("")
        data = self.c2.request_get(f"/api/v1/discovery/host/{self.id}")
        self._load_conf(data)

    def archive(self):
        """
        Archive the host.
        """
        if self.id is None:
            raise ExceptionTuoniDeleted("")
        self.c2.request_post("/api/v1/discovery/hosts/bulk-archive", {"hostIds": [self.id]})

    def update(self):
        """
        Update data on the server.
        """
        if self.id is None:
            raise ExceptionTuoniDeleted("")
        req = {
            "name": self.name,
            "note": self.note
        }
        self.c2.request_patch(f"/api/v1/discovery/host/{self.id}", req)
        self.reload()

